"""This module provides a CLI util to make updates to normalizer database."""
from os import environ
import logging
from timeit import default_timer as timer
from typing import List, Optional

import click
from botocore.exceptions import ClientError
from boto3.dynamodb.conditions import Key
from disease.database import Database as DiseaseDatabase
from disease.cli import CLI as DiseaseCLI
from disease.schemas import SourceName as DiseaseSources

from therapy import SOURCES_CLASS, SOURCES
from therapy.schemas import SourceName
from therapy.database import Database
from therapy.etl.merge import Merge

logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)


class CLI:
    """Class for updating the normalizer database via Click"""

    @staticmethod
    @click.command()
    @click.option(
        "--normalizer",
        help="The normalizer(s) you wish to update separated by spaces."
    )
    @click.option(
        "--prod",
        is_flag=True,
        help="Working in production environment."
    )
    @click.option(
        "--db_url",
        help="URL endpoint for the application database."
    )
    @click.option(
        "--update_all",
        is_flag=True,
        help="Update all normalizer sources."
    )
    @click.option(
        "--update_merged",
        is_flag=True,
        help="Update concepts for normalize endpoint from accepted sources."
    )
    @click.option(
        "--use_existing",
        is_flag=True,
        default=False,
        help="Use most recent existing source data instead of fetching latest version"
    )
    def update_normalizer_db(normalizer: str, prod: bool, db_url: str,
                             update_all: bool, update_merged: bool,
                             use_existing: bool) -> None:
        """Update selected normalizer source(s) in the therapy database.
        \f  # noqa: D301
        :param str normalizer: comma-separated string listing source names
        :param bool prod: if true, utilize production environment settings
        :param str db_url: DynamoDB endpoint URL (usually only needed locally)
        :param bool update_all: if true, update all sources
        :param bool update_merged: if true, update normalized group results
        :param bool use_existing: if true, don't try to fetch latest source data
        """
        endpoint_url = None
        if prod:
            environ["THERAPY_NORM_PROD"] = "TRUE"
            environ["DISEASE_NORM_PROD"] = "TRUE"
            db: Database = Database()
        else:
            if db_url:
                endpoint_url = db_url
            elif "THERAPY_NORM_DB_URL" in environ.keys():
                endpoint_url = environ["THERAPY_NORM_DB_URL"]
            else:
                endpoint_url = "http://localhost:8000"
            db = Database(db_url=endpoint_url)

        if update_all:
            normalizers = list(src for src in SOURCES)
            CLI()._check_disease_normalizer(normalizers, endpoint_url)
            CLI()._update_normalizers(normalizers, db, update_merged, use_existing)
        elif not normalizer:
            if update_merged:
                CLI()._update_merged(db, [])
            else:
                CLI()._help_msg()
        else:
            normalizers = str(normalizer).lower().split()

            if len(normalizers) == 0:
                CLI()._help_msg()

            non_sources = set(normalizers) - {src for src in SOURCES}

            if len(non_sources) != 0:
                raise Exception(f"Not valid source(s): {non_sources}")

            CLI()._check_disease_normalizer(normalizers, endpoint_url)
            CLI()._update_normalizers(normalizers, db, update_merged, use_existing)

    def _check_disease_normalizer(self, normalizers: List[str],
                                  endpoint_url: Optional[str]) -> None:
        """When loading HemOnc source, perform rudimentary check of Disease Normalizer
        tables, and reload them if necessary. This reload method should never be used
        (and is restricted from use) in a production setting.

        :param List[str] normalizers: List of sources to load
        :param Optional[str] endpoint_url: Therapy endpoint URL. This should always be
            a local address.
        """
        if "hemonc" in normalizers and "THERAPY_NORM_PROD" not in environ:
            db = DiseaseDatabase(db_url=endpoint_url)  # type: ignore
            current_tables = {table.name for table in db.dynamodb.tables.all()}
            if ("disease_concepts" not in current_tables) or \
                    ("disease_metadata" not in current_tables) or \
                    (db.diseases.scan()["Count"] == 0) or \
                    (db.metadata.scan()["Count"] < len(DiseaseSources)):

                msg = "Disease Normalizer not loaded. Loading now..."
                logger.debug(msg)
                click.echo(msg)
                try:
                    DiseaseCLI().update_normalizer_db(
                        ["--update_all", "--update_merged", "--db_url", endpoint_url]
                    )
                except Exception as e:
                    logger.error(e)
                    raise Exception(e)
                except:  # noqa: E722
                    # TODO: what does this do?
                    pass
                msg = "Disease Normalizer reloaded successfully."
                logger.debug(msg)
                click.echo(msg)

    @staticmethod
    def _help_msg() -> None:
        """Display help message."""
        ctx = click.get_current_context()
        click.echo(
            "Must either enter 1 or more sources, or use `--update_all` parameter")
        click.echo(ctx.get_help())
        ctx.exit()

    @staticmethod
    def _update_normalizers(normalizers: List[str], db: Database,
                            update_merged: bool, use_existing: bool) -> None:
        """Update selected normalizer sources.
        :param List[str] normalizers: list of source names to update
        :param Database db: database instance to use
        :param bool update_merged: if true, store concept IDs as they're processed and
            produce normalized records
        :param bool use_existing: if true, don't try to fetch latest source data in
            source perform_etl methods
        """
        processed_ids = list()
        for n in normalizers:
            msg = f"Deleting {n}..."
            click.echo(f"\n{msg}")
            logger.info(msg)

            start_delete = timer()
            CLI()._delete_data(n, db)
            end_delete = timer()
            delete_time = end_delete - start_delete

            msg = f"Deleted {n} in {delete_time:.5f} seconds."
            click.echo(f"{msg}\n")
            logger.info(msg)

            msg = f"Loading {n}..."
            click.echo(msg)
            logger.info(msg)

            start_load = timer()
            source = SOURCES_CLASS[n](database=db)
            try:
                processed_ids += source.perform_etl(use_existing)
            except FileNotFoundError as e:
                if use_existing:
                    if click.confirm(
                        f"Encountered FileNotFoundError while loading {n}: "
                        f"{e.args[0] if len(e.args) > 0 else ''}\n"
                        "Attempt to retrieve latest version from source? "
                    ):
                        processed_ids += source.perform_etl()
                    else:
                        raise e
            end_load = timer()
            load_time = end_load - start_load

            msg = f"Loaded {n} in {load_time:.5f} seconds."
            click.echo(msg)
            logger.info(msg)

            msg = f"Total time for {n}: " \
                  f"{(delete_time + load_time):.5f} seconds."
            click.echo(msg)
            logger.info(msg)

        if update_merged:
            CLI()._update_merged(db, processed_ids)

    def _update_merged(self, db: Database, processed_ids: List[str]) -> None:
        """Build and upload merged records. Will construct list of IDs if given an empty
        processed_ids list.
        :param Database db: DB instance to use
        :param List[str] processed_ids: List of IDs to create merged groups from
        """
        start_merge = timer()
        if not processed_ids:
            CLI()._delete_normalized_data(db)
            processed_ids = db.get_ids_for_merge()
        merge = Merge(database=db)
        click.echo("Constructing normalized records...")
        merge.create_merged_concepts(set(processed_ids))
        end_merge = timer()
        click.echo(f"Merged concept generation completed in"
                   f" {(end_merge - start_merge):.5f} seconds.")

    @staticmethod
    def _delete_normalized_data(database: Database) -> None:
        click.echo("\nDeleting normalized records...")
        start_delete = timer()
        try:
            while True:
                with database.therapies.batch_writer(
                        overwrite_by_pkeys=["label_and_type", "concept_id"]) \
                        as batch:
                    response = database.therapies.query(
                        IndexName="item_type_index",
                        KeyConditionExpression=Key("item_type").eq("merger"),
                    )
                    records = response["Items"]
                    if not records:
                        break
                    for record in records:
                        batch.delete_item(Key={
                            "label_and_type": record["label_and_type"],
                            "concept_id": record["concept_id"]
                        })
        except ClientError as e:
            click.echo(e.response["Error"]["Message"])
        end_delete = timer()
        delete_time = end_delete - start_delete
        click.echo(f"Deleted normalized records in {delete_time:.5f} seconds.")

    @staticmethod
    def _delete_data(source: str, database: Database) -> None:
        """Delete all data (records + metadata) from given source in database.
        :param str source: name of source to delete
        :param Database database: db instance
        """
        source_name = SourceName[f"{source.upper()}"].value
        # Delete source"s metadata first
        try:
            metadata = database.metadata.query(
                KeyConditionExpression=Key(
                    "src_name").eq(source_name)
            )
            if metadata["Items"]:
                database.metadata.delete_item(
                    Key={"src_name": metadata["Items"][0]["src_name"]},
                    ConditionExpression="src_name = :src",
                    ExpressionAttributeValues={
                        ":src": source_name}
                )
        except ClientError as e:
            click.echo(e.response["Error"]["Message"])

        try:
            while True:
                response = database.therapies.query(
                    IndexName="src_index",
                    KeyConditionExpression=Key("src_name").eq(
                        source_name),
                )

                records = response["Items"]
                if not records:
                    break

                with database.therapies.batch_writer(
                        overwrite_by_pkeys=["label_and_type", "concept_id"]) as batch:

                    for record in records:
                        batch.delete_item(
                            Key={
                                "label_and_type": record["label_and_type"],
                                "concept_id": record["concept_id"]
                            }
                        )
        except ClientError as e:
            click.echo(e.response["Error"]["Message"])


if __name__ == "__main__":
    CLI().update_normalizer_db()  # type: ignore



"""The VICC library for normalizing therapies."""
from pathlib import Path
import logging
import re

from .version import __version__

APP_ROOT: Path = Path(__file__).resolve().parents[0]
logging.basicConfig(
    filename="therapy.log",
    format="[%(asctime)s] - %(name)s - %(levelname)s : %(message)s")
logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)


class DownloadException(Exception):
    """Exception for failures relating to source file downloads."""

    def __init__(self, *args, **kwargs):  # noqa: ANN204
        """Initialize exception."""
        super().__init__(*args, **kwargs)


from therapy.schemas import SourceName, NamespacePrefix, ItemTypes  # noqa: E402, E501, I100, I202

# map plural to singular form
# eg {"label": "label", "trade_names": "trade_name"}
# key is the field name in the record object, value is the item_type value
# in reference objects
ITEM_TYPES = {k.lower(): v.value for k, v in ItemTypes.__members__.items()}

# Sources we import directly
SOURCES = {source.value.lower(): source.value
           for source in SourceName.__members__.values()}

# use to fetch source name from schema based on concept id namespace
# e.g. {'chembl': 'ChEMBL'}
PREFIX_LOOKUP = {v.value: SourceName[k].value
                 for k, v in NamespacePrefix.__members__.items()
                 if k in SourceName.__members__.keys()}

# Namespace LUI patterns. Use for namespace inference.
NAMESPACE_LUIS = (
    (re.compile(r"^CHEMBL\d+$", re.IGNORECASE), SourceName.CHEMBL.value),
    (re.compile(r"^\d+\-\d+\-\d+$", re.IGNORECASE), SourceName.CHEMIDPLUS.value),
    (re.compile(r"^(Q|P)\d+$", re.IGNORECASE), SourceName.WIKIDATA.value),
    (re.compile(r"^C\d+$", re.IGNORECASE), SourceName.NCIT.value),
    (re.compile(r"^DB\d{5}$", re.IGNORECASE), SourceName.DRUGBANK.value),
    (re.compile(r"^(A?NDA)(\d+)$", re.IGNORECASE), SourceName.DRUGSATFDA.value)
)

# Sources that we import directly
XREF_SOURCES = {source for source in SourceName.__members__}

# Sources that are found in data from imported sources
ASSOC_WITH_SOURCES = {source for source in NamespacePrefix.__members__} - XREF_SOURCES

from therapy.etl import ChEMBL, Wikidata, DrugBank, NCIt, ChemIDplus, RxNorm, HemOnc, GuideToPHARMACOLOGY, DrugsAtFDA  # noqa: E402, E501, I202

# used to get source class name from string
SOURCES_CLASS = \
    {s.value.lower(): eval(s.value) for s in SourceName.__members__.values()}



"""This module defines the RxNorm ETL methods.

"This product uses publicly available data courtesy of the U.S. National
Library of Medicine (NLM), National Institutes of Health, Department of Health
 and Human Services; NLM is not responsible for the product and does not
 endorse or recommend this or any other product."
"""
import csv
import logging
import shutil
import zipfile
import re
from os import environ, remove
from typing import List, Dict
from pathlib import Path

import yaml
import bioversions
from boto3.dynamodb.table import BatchWriter
import requests

from therapy import DownloadException, XREF_SOURCES, ASSOC_WITH_SOURCES, ITEM_TYPES
from therapy.schemas import SourceName, NamespacePrefix, SourceMeta, Drug, \
    ApprovalRating, RecordParams
from therapy.etl.base import Base

logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)

# Designated Alias, Designated Syn, Tall Man Syn, Machine permutation
# Generic Drug Name, Designated Preferred Name, Preferred Entry Term,
# Clinical Drug, Entry Term, Rxnorm Preferred
ALIASES = ["SYN", "SY", "TMSY", "PM", "GN", "PT", "PEP", "CD", "ET", "RXN_PT"]

# Fully-specified drug brand name that can be prescribed
# Fully-specified drug brand name that can not be prescribed,
# Semantic branded drug
TRADE_NAMES = ["BD", "BN", "SBD"]

# Allowed rxnorm xrefs that have Source Level Restriction 0 or 1
RXNORM_XREFS = ["ATC", "CVX", "DRUGBANK", "MMSL", "MSH", "MTHCMSFRF", "MTHSPL",
                "RXNORM", "USP", "VANDF"]


class RxNorm(Base):
    """Class for RxNorm ETL methods."""

    def _create_drug_form_yaml(self) -> None:
        """Create a YAML file containing RxNorm drug form values."""
        self._drug_forms_file = self._src_dir / f"rxnorm_drug_forms_{self._version}.yaml"  # noqa: E501
        dfs = []
        with open(self._src_file) as f:  # type: ignore
            data = csv.reader(f, delimiter="|")
            for row in data:
                if row[12] == "DF" and row[11] == "RXNORM":
                    if row[14] not in dfs:
                        dfs.append(row[14])
        with open(self._drug_forms_file, "w") as file:
            yaml.dump(dfs, file)

    def _zip_handler(self, dl_path: Path, outfile_path: Path) -> None:
        """Extract required files from RxNorm zip. This method should be passed to
        the base class's _http_download method.
        :param Path dl_path: path to RxNorm zip file in tmp directory
        :param Path outfile_path: path to RxNorm data directory
        """
        rrf_path = outfile_path / f"rxnorm_{self._version}.RRF"
        with zipfile.ZipFile(dl_path, "r") as zf:
            rrf = zf.open("rrf/RXNCONSO.RRF")
            target = open(rrf_path, "wb")
            with rrf, target:
                shutil.copyfileobj(rrf, target)
        remove(dl_path)
        self._src_file = rrf_path
        self._create_drug_form_yaml()
        logger.info("Successfully retrieved source data for RxNorm")

    def _download_data(self) -> None:
        """Download latest RxNorm data file.

        :raises DownloadException: if API Key is not defined in the environment.
        """
        logger.info("Retrieving source data for RxNorm")
        api_key = environ.get("RXNORM_API_KEY")
        if api_key is None:
            logger.error("Could not find RXNORM_API_KEY in environment variables.")
            raise DownloadException("RXNORM_API_KEY not found.")

        url = bioversions.resolve("rxnorm").homepage
        if not url:
            raise DownloadException("Could not resolve RxNorm homepage")

        tgt_data = {"apikey": api_key}
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        api_url = "https://utslogin.nlm.nih.gov/cas/v1/api-key"
        tgt_r = requests.post(api_url,
                              data=tgt_data, headers=headers)
        tgt_matches = re.findall(r'https://.+(TGT.+)" m', tgt_r.text)
        if not tgt_matches:
            raise DownloadException("Unable to retrieve TGT")
        tgt_value = tgt_matches[0]

        st_data = {"service": url}
        st_url = f"https://utslogin.nlm.nih.gov/cas/v1/tickets/{tgt_value}"
        st_r = requests.post(st_url, data=st_data, headers=headers)

        self._http_download(f"{url}?ticket={st_r.text}", self._src_dir,
                            handler=self._zip_handler)

    def _extract_data(self, use_existing: bool = False) -> None:
        """Get source files from RxNorm data directory.
        This class expects a file named `rxnorm_<version>.RRF` and a file named
        `rxnorm_drug_forms_<version>.yaml`. This method will download and
        generate them if they are unavailable.

        :param bool use_existing: if True, don't try to fetch latest source data
        """
        super()._extract_data(use_existing)
        drug_forms_path = self._src_dir / f"rxnorm_drug_forms_{self._version}.yaml"
        if not drug_forms_path.exists():
            self._create_drug_form_yaml()
        else:
            self._drug_forms_file = drug_forms_path

    def _transform_data(self) -> None:
        """Transform the RxNorm source."""
        with open(self._drug_forms_file, "r") as file:
            drug_forms = yaml.safe_load(file)

        with open(self._src_file) as f:
            rff_data = csv.reader(f, delimiter="|")
            # Link ingredient to brand
            ingredient_brands: Dict[str, str] = dict()
            # Link precise ingredient to get brand
            precise_ingredient: Dict[str, str] = dict()
            # Transformed therapy records
            data: Dict[str, Dict] = dict()
            # Link ingredient to brand
            sbdfs: Dict[str, str] = dict()
            # Get RXNORM|BN to concept_id
            brands: Dict[str, str] = dict()
            for row in rff_data:
                if row[11] in RXNORM_XREFS:
                    concept_id = f"{NamespacePrefix.RXNORM.value}:{row[0]}"
                    if row[12] == "BN" and row[11] == "RXNORM":
                        brands[row[14]] = concept_id
                    if row[12] == "SBDC" and row[11] == "RXNORM":
                        # Semantic Branded Drug Component
                        self._get_brands(row, ingredient_brands)
                    else:
                        if concept_id not in data.keys():
                            params: RecordParams = {}
                            params["concept_id"] = concept_id
                            self._add_str_field(params, row, precise_ingredient,
                                                drug_forms, sbdfs)
                            self._add_xref_assoc(params, row)
                            data[concept_id] = params
                        else:
                            # Concept already created
                            params = data[concept_id]
                            self._add_str_field(params, row, precise_ingredient,
                                                drug_forms, sbdfs)
                            self._add_xref_assoc(params, row)

            with self.database.therapies.batch_writer() as batch:
                for value in data.values():
                    if "label" in value:
                        self._get_trade_names(value, precise_ingredient,
                                              ingredient_brands, sbdfs)
                        self._load_brand_concepts(value, brands, batch)

                        params = {"concept_id": value["concept_id"]}

                        for field in list(ITEM_TYPES.keys()) + ["approval_ratings"]:
                            field_value = value.get(field)
                            if field_value:
                                params[field] = field_value
                        assert Drug(**params)
                        self._load_therapy(params)

    def _get_brands(self, row: List, ingredient_brands: Dict) -> None:
        """Add ingredient and brand to ingredient_brands.

        :param List row: A row in the RxNorm data file.
        :param Dict ingredient_brands: Store brands for each ingredient
        """
        # SBDC: Ingredient(s) + Strength + [Brand Name]
        term = row[14]
        ingredients_brand = \
            re.sub(r"(\d*)(\d*\.)?\d+ (MG|UNT|ML)?(/(ML|HR|MG))?",
                   "", term)
        brand = term.split("[")[-1].split("]")[0]
        ingredients = ingredients_brand.replace(f"[{brand}]", "")
        if "/" in ingredients:
            ingredients = ingredients.split("/")
            for ingredient in ingredients:
                self._add_term(ingredient_brands, brand,
                               ingredient.strip())
        else:
            self._add_term(ingredient_brands, brand,
                           ingredients.strip())

    def _get_trade_names(self, value: Dict, precise_ingredient: Dict,
                         ingredient_brands: Dict, sbdfs: Dict) -> None:
        """Get trade names for a given ingredient.

        :param Dict value: Therapy attributes
        :param Dict precise_ingredient: Brand names for precise ingredient
        :param Dict ingredient_brands: Brand names for ingredient
        :param Dict sbdfs: Brand names for ingredient from SBDF row
        """
        record_label = value["label"].lower()
        labels = [record_label]

        if "PIN" in value and value["PIN"] \
                in precise_ingredient:
            for pin in precise_ingredient[value["PIN"]]:
                labels.append(pin.lower())

        for label in labels:
            trade_names: List[str] = \
                [val for key, val in ingredient_brands.items()
                 if label == key.lower()]
            trade_names_uq = {val for sublist in trade_names
                              for val in sublist}
            for tn in trade_names_uq:
                self._add_term(value, tn, "trade_names")

        if record_label in sbdfs:
            for tn in sbdfs[record_label]:
                self._add_term(value, tn, "trade_names")

    @staticmethod
    def _load_brand_concepts(value: Dict, brands: Dict, batch: BatchWriter) -> None:
        """Connect brand names to a concept and load into the database.

        :params dict value: A transformed therapy record
        :params dict brands: Connects brand names to concept records
        :param BatchWriter batch: Object to write data to DynamoDB.
        """
        if "trade_names" in value:
            for tn in value["trade_names"]:
                if brands.get(tn):
                    batch.put_item(Item={
                        "label_and_type":
                            f"{brands.get(tn)}##rx_brand",
                        "concept_id": value["concept_id"],
                        "src_name": SourceName.RXNORM.value,
                        "item_type": "rx_brand"
                    })

    def _add_str_field(self, params: Dict, row: List, precise_ingredient: Dict,
                       drug_forms: List, sbdfs: Dict) -> None:
        """Differentiate STR field.

        :param Dict params: A transformed therapy record.
        :param List row: A row in the RxNorm data file.
        :param Dict precise_ingredient: Precise ingredient information
        :param List drug_forms: RxNorm Drug Form values
        :param Dict sbdfs: Brand names for precise ingredient
        """
        term = row[14]
        term_type = row[12]
        source = row[11]

        if (term_type == "IN" or term_type == "PIN") and source == "RXNORM":
            params["label"] = term
            if row[17] == "4096":
                params["approval_ratings"] = [ApprovalRating.RXNORM_PRESCRIBABLE.value]
        elif term_type in ALIASES:
            self._add_term(params, term, "aliases")
        elif term_type in TRADE_NAMES:
            self._add_term(params, term, "trade_names")

        if source == "RXNORM":
            if term_type == "SBDF":
                brand = term.split("[")[-1].split("]")[0]
                ingredient_strength = term.replace(f"[{brand}]", "")
                for df in drug_forms:
                    if df in ingredient_strength:
                        ingredient = \
                            ingredient_strength.replace(df, "").strip()
                        self._add_term(sbdfs, brand, ingredient.lower())
                        break

        if source == "MSH":
            if term_type == "MH":
                # Get ID for accessing precise ingredient
                params["PIN"] = row[13]
            elif term_type == "PEP":
                self._add_term(precise_ingredient, term, row[13])

    @staticmethod
    def _add_term(params: Dict, term: str, label_type: str) -> None:
        """Add a single term to a therapy record in an associated field.

        :param Dict params: A transformed therapy record.
        :param Str term: The term to be added
        :param Str label_type: The type of term
        """
        if label_type in params and params[label_type]:
            if term not in params[label_type]:
                params[label_type].append(term)
        else:
            params[label_type] = [term]

    def _add_xref_assoc(self, params: Dict, row: List) -> None:
        """Add xref or associated_with to therapy.

        :param Dict params: A transformed therapy record.
        :param List row: A row in the RxNorm data file.
        """
        ref = row[11]
        if ref:
            if ref == "MTHSPL":
                xref_assoc = "UNII"
            else:
                xref_assoc = row[11].upper()

            if xref_assoc in XREF_SOURCES:
                source_id =\
                    f"{NamespacePrefix[xref_assoc].value}:{row[13]}"
                if source_id != params["concept_id"]:
                    # Sometimes concept_id is included in the source field
                    self._add_term(params, source_id, "xrefs")
            elif xref_assoc in ASSOC_WITH_SOURCES:
                source_id = f"{NamespacePrefix[xref_assoc].value}:{row[13]}"
                self._add_term(params, source_id, "associated_with")
            else:
                logger.info(f"{xref_assoc} not in NameSpacePrefix.")

    def _load_meta(self) -> None:
        """Add RxNorm metadata."""
        meta = SourceMeta(
            data_license="UMLS Metathesaurus",
            data_license_url="https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html",  # noqa: E501
            version=self._version,
            data_url=bioversions.resolve("rxnorm").homepage,
            rdp_url=None,
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": True
            }
        )
        params = dict(meta)
        params["src_name"] = SourceName.RXNORM.value
        self.database.metadata.put_item(Item=params)



"""ETL methods for ChemIDPlus® source.

Courtesy of the U.S. National Library of Medicine.
"""
import logging
import zipfile
from os import remove
from shutil import move
from typing import Generator
from pathlib import Path
import xml.etree.ElementTree as ET
import re

from therapy.etl.base import Base
from therapy.schemas import NamespacePrefix, SourceMeta, SourceName, \
    DataLicenseAttributes, RecordParams


logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)


TAGS_REGEX = r" \[.*\]"


class ChemIDplus(Base):
    """Class for ChemIDplus ETL methods."""

    def _download_data(self) -> None:
        """Download source data from default location."""
        logger.info("Retrieving source data for ChemIDplus")
        file = "currentchemid.zip"
        self._ftp_download("ftp.nlm.nih.gov", "nlmdata/.chemidlease", file)
        zip_path = (self._src_dir / file).absolute()
        zip_file = zipfile.ZipFile(zip_path, "r")
        outfile = self._src_dir / f"chemidplus_{self._version}.xml"
        for info in zip_file.infolist():
            if re.match(r".*\.xml", info.filename):
                xml_filename = info.filename
                zip_file.extract(info, path=self._src_dir)
                move(str(self._src_dir / xml_filename), outfile)
                break
        remove(zip_path)
        assert outfile.exists()
        logger.info("Successfully retrieved source data for ChemIDplus")

    @staticmethod
    def parse_xml(path: Path, tag: str) -> Generator:
        """Parse XML file and retrieve elements with matching tag value.
        :param Path path: path to XML file
        :param str tag: XML tag
        :return: generator yielding elements of corresponding tag
        """
        context = iter(ET.iterparse(path, events=("start", "end")))
        _, root = next(context)
        for event, elem in context:
            if event == "end" and elem.tag == tag:
                yield elem
                root.clear()

    def _transform_data(self) -> None:
        """Open dataset and prepare for loading into database."""
        parser = self.parse_xml(self._src_file, "Chemical")
        for chemical in parser:  # type: ignore
            if "displayName" not in chemical.attrib:
                continue

            # initial setup and get label
            display_name = chemical.attrib["displayName"]
            if not display_name or not re.search(TAGS_REGEX, display_name):
                continue
            label = re.sub(TAGS_REGEX, "", display_name)
            params: RecordParams = {"label": label}

            # get concept ID
            reg_no = chemical.find("NumberList").find("CASRegistryNumber")
            if not reg_no:
                continue
            params["concept_id"] = f"{NamespacePrefix.CASREGISTRY.value}:{reg_no.text}"

            # get aliases
            aliases = []
            label_l = label.lower()
            name_list = chemical.find("NameList")
            if name_list:
                for name in name_list.findall("NameOfSubstance"):
                    text = name.text
                    if text != display_name and text.lower() != label_l:
                        aliases.append(re.sub(TAGS_REGEX, "", text))
            params["aliases"] = aliases

            # get xrefs and associated_with
            params["xrefs"] = []
            params["associated_with"] = []
            locator_list = chemical.find("LocatorList")
            if locator_list:
                for loc in locator_list.findall("InternetLocator"):
                    if loc.text == "DrugBank":
                        db = f"{NamespacePrefix.DRUGBANK.value}:" \
                             f"{loc.attrib['url'].split('/')[-1]}"
                        params["xrefs"].append(db)  # type: ignore
                    elif loc.text == "FDA SRS":
                        unii = f"{NamespacePrefix.UNII.value}:" \
                               f"{loc.attrib['url'].split('/')[-1]}"
                        params["associated_with"].append(unii)  # type: ignore

            self._load_therapy(params)

    def _load_meta(self) -> None:
        """Add source metadata."""
        meta = SourceMeta(
            data_license="custom",
            data_license_url="https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",  # noqa: E501
            version=self._version,
            data_url="ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",
            rdp_url=None,
            data_license_attributes=DataLicenseAttributes(
                non_commercial=False,
                share_alike=False,
                attribution=True
            )
        )
        item = dict(meta)
        item["src_name"] = SourceName.CHEMIDPLUS.value
        self.database.metadata.put_item(Item=item)



"""This module defines the DrugBank ETL methods."""
from typing import Dict, Any
import logging
import csv

from therapy.schemas import SourceName, SourceMeta, NamespacePrefix
from therapy.etl.base import Base

logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)


class DrugBank(Base):
    """Class for DrugBank ETL methods."""

    def _download_data(self) -> None:
        """Download DrugBank source data."""
        logger.info("Retrieving source data for DrugBank")
        url = f"https://go.drugbank.com/releases/{self._version.replace('.', '-')}/downloads/all-drugbank-vocabulary"  # noqa: E501
        csv_file = self._src_dir / f"drugbank_{self._version}.csv"
        self._http_download(url, csv_file, handler=self._zip_handler)
        logger.info("Successfully retrieved source data for DrugBank")

    def _load_meta(self) -> None:
        """Add DrugBank metadata."""
        meta = {
            "data_license": "CC0 1.0",
            "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/",
            "version": self._version,
            "data_url": "https://go.drugbank.com/releases/latest#open-data",
            "rdp_url": "http://reusabledata.org/drugbank.html",
            "data_license_attributes": {
                "non_commercial": False,
                "share_alike": False,
                "attribution": False,
            },
        }
        assert SourceMeta(**meta)
        meta["src_name"] = SourceName.DRUGBANK.value
        self.database.metadata.put_item(Item=meta)

    def _transform_data(self) -> None:
        """Transform the DrugBank source."""
        with open(self._src_file, "r") as file:
            reader = csv.reader(file)
            next(reader)  # skip header
            for row in reader:
                # get concept ID
                params: Dict[str, Any] = {
                    "concept_id": f"{NamespacePrefix.DRUGBANK.value}:{row[0]}",
                }

                # get label
                label = row[2]
                if label:
                    params["label"] = label

                # get aliases
                aliases = [a for a in row[1].split(" | ") + row[5].split(" | ") if a]
                if aliases:
                    params["aliases"] = aliases

                # get CAS reference
                cas_ref = row[3]
                if cas_ref:
                    params["xrefs"] = [f"{NamespacePrefix.CHEMIDPLUS.value}:{row[3]}"]

                params["associated_with"] = []
                # get inchi key
                if len(row) >= 7:
                    inchi_key = row[6]
                    if inchi_key:
                        inchi_id = f"{NamespacePrefix.INCHIKEY.value}:{inchi_key}"
                        params["associated_with"].append(inchi_id)
                # get UNII id
                unii = row[4]
                if unii:
                    unii_id = f"{NamespacePrefix.UNII.value}:{unii}"
                    params["associated_with"].append(unii_id)

                self._load_therapy(params)
